@@ -6,11 +6,15 @@
 
 import numpy as np
 
-import fio, kine, kine_or, vtkh, ligaments as liga, contacts
+from . import fio, mplh, kine, kine_or, vtkh, ligaments as liga, contacts, spine
 
 import re
 from itertools import groupby
 
+import os
+from mpl_toolkits.mplot3d import Axes3D
+import matplotlib.pyplot as plt
+
 
 def expressOptoWandTipToMimicsRefFrame(
                                         filePathC3D, 
@@ -135,7 +139,7 @@ def expressOptoWandTipToMimicsRefFrame(
     args = {}
     args['mkrsLoc'] = markersMimics
     args['verbose'] = verbose
-    oRm, oTm = kine.rigidBodySVDFun(markersOpto, mkrList, args)
+    oRm, oTm, info = kine.rigidBodySVDFun(markersOpto, mkrList, args)
     oRTm = kine.composeRotoTranslMatrix(oRm, oTm)
     
     # Calculate pose from Optoelectronic reference frame to Mimics reference frame
@@ -234,37 +238,55 @@ def calculateKneeSegmentsPoses(
     args['verbose'] = verbose
     
     femurAvailable = False
+    femurError = False
     mkrList = ["mF1", "mF2", "mF3", "mF4"]
     if set(mkrList) <= set(markersLoc.keys()):
         femurAvailable = True
         print('==== ---- Femur markers found')
-        R1, T1, info1 = kine.rigidBodySVDFun(markersOpto, mkrList, args)
-        RT1 = kine.composeRotoTranslMatrix(R1, T1)
+        try:
+            R1, T1, info1 = kine.rigidBodySVDFun(markersOpto, mkrList, args)
+            RT1 = kine.composeRotoTranslMatrix(R1, T1)
+        except Exception as e:
+            print('Impossible to calculate femur pose')
+            print(e)
+            femurError = True
     
     tibiaAvailable = False
+    tibiaError = False
     mkrList = ["mT1", "mT2", "mT3", "mT4"]
     if set(mkrList) <= set(markersLoc.keys()):
         tibiaAvailable = True
         print('==== ---- Tibia markers found')
-        R2, T2, info2 = kine.rigidBodySVDFun(markersOpto, mkrList, args)
-        RT2 = kine.composeRotoTranslMatrix(R2, T2)
+        try:
+            R2, T2, info2 = kine.rigidBodySVDFun(markersOpto, mkrList, args)
+            RT2 = kine.composeRotoTranslMatrix(R2, T2)
+        except Exception as e:
+            print('Impossible to calculate tibia pose')
+            print(e)
+            tibiaError = True
     
     patellaAvailable = False
+    patellaError = False
     mkrList = ["mP1", "mP2", "mP3", "mP4"]
     if set(mkrList) <= set(markersLoc.keys()):
         patellaAvailable = True
         print('==== ---- Patella markers found')
-        R3, T3, info3 = kine.rigidBodySVDFun(markersOpto, mkrList, args)
-        RT3 = kine.composeRotoTranslMatrix(R3, T3)
+        try:
+            R3, T3, info3 = kine.rigidBodySVDFun(markersOpto, mkrList, args)
+            RT3 = kine.composeRotoTranslMatrix(R3, T3)
+        except Exception as e:
+            print('Impossible to calculate tibia pose')
+            print(e)
+            patellaError = True
     
     results = {}
-    if femurAvailable:
+    if femurAvailable and not femurError:
         results['femur_pose'] = RT1
         results['femur_pose_reconstruction_info'] = info1
-    if tibiaAvailable:
+    if tibiaAvailable and not tibiaError:
         results['tibia_pose'] = RT2
         results['tibia_pose_reconstruction_info'] = info2
-    if patellaAvailable:
+    if patellaAvailable and not patellaError:
         results['patella_pose'] = RT3
         results['patella_pose_reconstruction_info'] = info3
     
@@ -384,6 +406,7 @@ def calculateKneeLigamentsData(
                                 frames = None,
                                 ligaNames = ['MCL', 'LCL'],
                                 tibiaPlateauMedEdgeSplineLoc = None,
+                                tibiaPlateauLatEdgeSplineLoc = None,
                                 ligaModels = ['straight','Blankevoort_1991'],
                                 vtkFemur = None,
                                 vtkTibia = None,
@@ -398,6 +421,12 @@ def calculateKneeLigamentsData(
         tibiaPlateauMedEdgeParamsSplineLoc = vtkh.createParamSpline(tibiaPlateauMedEdgeSplineLoc)
     else:
         tibiaPlateauMedEdgeParamsSplineLoc = None
+        
+    print('==== Creating spline for tibial lateral edge plateau ...')
+    if tibiaPlateauLatEdgeSplineLoc is not None:
+        tibiaPlateauLatEdgeParamsSplineLoc = vtkh.createParamSpline(tibiaPlateauLatEdgeSplineLoc)
+    else:
+        tibiaPlateauLatEdgeParamsSplineLoc = None
     
     # initialize all insertion points dict
     allPointsIns = {}
@@ -413,12 +442,48 @@ def calculateKneeLigamentsData(
             usedIns2.append("TMCL")
             availableLigaNames.append("MCL")
             depsIns['MCL'] = ['FMCL', 'TMCL']
+    if "MCLa" in ligaNames:
+        if "FMCLa" in insertionsLoc and "TMCLa" in insertionsLoc:
+            usedIns1.append("FMCLa")
+            usedIns2.append("TMCLa")
+            availableLigaNames.append("MCLa")
+            depsIns['MCLa'] = ['FMCLa', 'TMCLa']
+    if "MCLm" in ligaNames:
+        if "FMCLm" in insertionsLoc and "TMCLm" in insertionsLoc:
+            usedIns1.append("FMCLm")
+            usedIns2.append("TMCLm")
+            availableLigaNames.append("MCLm")
+            depsIns['MCLm'] = ['FMCLm', 'TMCLm']
+    if "MCLp" in ligaNames:
+        if "FMCLp" in insertionsLoc and "TMCLp" in insertionsLoc:
+            usedIns1.append("FMCLp")
+            usedIns2.append("TMCLp")
+            availableLigaNames.append("MCLp")
+            depsIns['MCLp'] = ['FMCLp', 'TMCLp']
     if "LCL" in ligaNames:
         if "FLCL" in insertionsLoc and "TLCL" in insertionsLoc:
             usedIns1.append("FLCL")
             usedIns2.append("TLCL")
             availableLigaNames.append("LCL")
             depsIns['LCL'] = ['FLCL', 'TLCL']
+    if "LCLa" in ligaNames:
+        if "FLCLa" in insertionsLoc and "TLCLa" in insertionsLoc:
+            usedIns1.append("FLCLa")
+            usedIns2.append("TLCLa")
+            availableLigaNames.append("LCLa")
+            depsIns['LCLa'] = ['FLCLa', 'TLCLa']
+    if "LCLm" in ligaNames:
+        if "FLCLm" in insertionsLoc and "TLCLm" in insertionsLoc:
+            usedIns1.append("FLCLm")
+            usedIns2.append("TLCLm")
+            availableLigaNames.append("LCLm")
+            depsIns['LCLm'] = ['FLCLm', 'TLCLm']
+    if "LCLp" in ligaNames:
+        if "FLCLp" in insertionsLoc and "TLCLp" in insertionsLoc:
+            usedIns1.append("FLCLp")
+            usedIns2.append("TLCLp")
+            availableLigaNames.append("LCLp")
+            depsIns['LCLp'] = ['FLCLp', 'TLCLp']
 
     # express FEMUR insertion points in lab reference frame
     print('==== Expressing femur ligament insertions in Optoelectronic reference frame ...')
@@ -492,25 +557,37 @@ def calculateKneeLigamentsData(
                     
                 if 'Blankevoort_1991' not in ligaLengths[ligaName]:
                     ligaLengths[ligaName]['Blankevoort_1991'] = Nf * [np.nan]
-    
-                if tibiaPlateauMedEdgeParamsSplineLoc is not None:
-        
+                    
+                if 'MCL' in ligaName:
+                    
+                    if tibiaPlateauMedEdgeParamsSplineLoc is None:
+                    
+                        raise Exception('"Blankevoort_1991" methods requires input "tibiaPlateauMedEdgeSplineLoc" for MCL(a,m,p)')
+                        
                     # repose spline
                     tibiaPlateauMedEdgeParamsSpline = vtkh.reposeSpline(tibiaPlateauMedEdgeParamsSplineLoc, RT2[i,...])
-        
+    
                     # 2-lines model with shortest possible path if touching tibial edge
                     dummy, ligaPathB = liga.ligamentPathBlankevoort1991(pIns2, tibiaPlateauMedEdgeParamsSpline)
                     
-                    # Calculate length
-                    ligaLengthB = liga.calcLigaLength(ligaPathB)
+                elif 'LCL' in ligaName:
                     
-                    ligaPaths[ligaName]['Blankevoort_1991'][i] = ligaPathB
+                    if tibiaPlateauLatEdgeParamsSplineLoc is None:
+                    
+                        raise Exception('"Blankevoort_1991" methods requires input "tibiaPlateauLatEdgeSplineLoc" for LCL(a,m,p)')
+                        
+                    # repose spline
+                    tibiaPlateauLatEdgeParamsSpline = vtkh.reposeSpline(tibiaPlateauLatEdgeParamsSplineLoc, RT2[i,...])
+    
+                    # 2-lines model with shortest possible path if touching tibial edge
+                    dummy, ligaPathB = liga.ligamentPathBlankevoort1991(pIns2, tibiaPlateauLatEdgeParamsSpline)
                 
-                    ligaLengths[ligaName]['Blankevoort_1991'][i] = ligaLengthB
-                    
-                else:
-                    
-                    raise Exception('"Blankevoort_1991" methods requires input "tibiaPlateauMedEdgeSplineLoc"')
+                # Calculate length
+                ligaLengthB = liga.calcLigaLength(ligaPathB)
+                
+                ligaPaths[ligaName]['Blankevoort_1991'][i] = ligaPathB
+            
+                ligaLengths[ligaName]['Blankevoort_1991'][i] = ligaLengthB
                     
             if 'Marai_2004' in ligaModels:
                 
@@ -560,6 +637,15 @@ def calculateKneeLigamentsData(
                 
                 actors.append(vtkh.createVTKActor(vtkTibiaPlateauMedEdgeLine))
                 names.append('tibia_medial_plateau')
+
+            if tibiaPlateauLatEdgeParamsSplineLoc is not None:
+            
+                tibiaPlateauLatEdgePoints = vtkh.evalSpline(tibiaPlateauLatEdgeParamsSpline, np.arange(0, 1.01, 0.01))
+                
+                vtkTibiaPlateauLatEdgeLine = vtkh.createLineVTKData(tibiaPlateauLatEdgePoints, [255, 0, 0])
+                
+                actors.append(vtkh.createVTKActor(vtkTibiaPlateauLatEdgeLine))
+                names.append('tibia_lateral_plateau')
             
             for ligaName in ligaPaths:
                 
@@ -723,7 +809,7 @@ def assembleKneeDataGranular(
     
     data = Nf * [None]
     
-    for i in xrange(Nf):
+    for i in range(Nf):
         
         row = []
         
@@ -841,4 +927,233 @@ def assembleKneeDataAsIsNoMetadata(
     
     return data
     
+    
+    
+    
+
+def performSpineAnalysis(
+                        markers1,
+                        markers2,
+                        singleMarkersDescFile,
+                        segmentsDescFile,
+                        clustersDescDir,
+                        spinePointNames,
+                        resultsDir,
+                        savePlots=True
+                        ):
+                            
+    # calculate THORAX anatomical reference frame for acquisition system 1
+    markers1Tho = {}
+    markers1Tho['IJ'] = markers1['CLAV']
+    markers1Tho['PX'] = markers1['STRN']
+    markers1Tho['C7'] = markers1['C7']
+    markers1Tho['T8'] = markers1['T9']
+    RTho1, OTho1 = kine.thoraxPoseISB(markers1Tho)
+    RTTho1 = kine.composeRotoTranslMatrix(RTho1, OTho1)
+    RTTho1i = kine.inv2(RTTho1)
+    
+    # calculate PELVIS anatomical reference frame for acquisition system 1
+    RPel1, OPel1 = kine.pelvisPoseISB(markers1, s='R')
+    RTPel1 = kine.composeRotoTranslMatrix(RPel1, OPel1)
+    RTPel1i = kine.inv2(RTPel1)
+    
+    # calculate THORAX anatomical reference frame for acquisition system 2
+    markers2Tho = {}
+    markers2Tho['IJ'] = markers2['CLAV']
+    markers2Tho['PX'] = markers2['STRN']
+    markers2Tho['C7'] = markers2['C7']
+    markers2Tho['T8'] = markers2['T9']
+    RTho2, OTho2 = kine.thoraxPoseISB(markers2Tho)
+    RTTho2 = kine.composeRotoTranslMatrix(RTho2, OTho2)
+    
+    # calculate PELVIS anatomical reference frame for acquisition system 2
+    RPel2, OPel2 = kine.pelvisPoseISB(markers2, s='R')
+    RTPel2 = kine.composeRotoTranslMatrix(RPel2, OPel2)
+    RTPel2i = kine.inv2(RTPel2)
+    
+    # Handles single markers
+    markers2New = {}
+    singleMarkersInfo = fio.readStringListMapFile(singleMarkersDescFile)
+    for m in singleMarkersInfo:
+        markerName = m
+        segment = singleMarkersInfo[m][0]
+        
+        # Express real marker in anatomical reference frame 
+        if segment == 'thorax':
+            RT1i = RTTho1i
+        elif segment == 'pelvis':
+            RT1i = RTPel1i
+        else:
+            raise Exception('segment must be one of: thorax, pelvis')
+        targetMarkers = {}
+        targetMarkers[markerName + ' base'] = markers1[markerName + ' base']
+        targetMarkers['True ' + markerName] = markers1['True ' + markerName]
+        targetMarkersLoc = kine.changeMarkersReferenceFrame(targetMarkers, RT1i)
+        
+        #
+        if segment == 'thorax':
+            RT2 = RTTho2
+        elif segment == 'pelvis':
+            RT2 = RTPel2
+        else:
+            raise Exception('segment must be one of: thorax, pelvis')
+        markers2New.update(kine.changeMarkersReferenceFrame(targetMarkersLoc, RT2))
+        
+    
+    # Handles clusters
+    segmentsInfo = fio.readStringListMapFile(segmentsDescFile)
+    print(segmentsInfo)
+    
+    for s in segmentsInfo:
+        segmentName = s
+        clusterType = segmentsInfo[s][0]
+        print('%s %s' % (segmentName, clusterType))
+        
+        # Read data for cluster connected to current segment
+        clusterInfo = fio.readStringListMapFile(os.path.join(clustersDescDir, clusterType + '.txt'))
+        clusterInfo = {m: np.array(clusterInfo[m]) for m in clusterInfo}
+        
+        # Modify cluster marker names with real ones
+        clusterInfoSpec = clusterInfo.copy()
+        subs = segmentsInfo[s][1:]
+        clusterBaseName = clusterBaseNameSpec = 'Base'
+        for sub in subs:
+            s = sub.split(':')
+            if s[0] == clusterBaseName:
+                clusterBaseNameSpec = s[1]
+            del clusterInfoSpec[s[0]]
+            clusterInfoSpec[s[1]] = clusterInfo[s[0]]
+        print(clusterInfoSpec)
+        
+        # SVD for acquisition system 1
+        args = {}
+        markersLoc = clusterInfoSpec.copy()
+        args['mkrsLoc'] = markersLoc
+        args['verbose'] = True
+        mkrList = markersLoc.keys()
+        R1, T1, infoSVD1 = kine.rigidBodySVDFun(markers1, mkrList, args)
+        RT1 = kine.composeRotoTranslMatrix(R1, T1)
+        
+        # Express cluster base and real marker in cluster reference frame
+        RT1i = kine.inv2(RT1)
+        targetMarkers = {}
+        targetMarkers[clusterBaseNameSpec + ' 1'] = markers1[clusterBaseNameSpec]
+        targetMarkers['True ' + segmentName] = markers1['True ' + segmentName]
+        targetMarkersLoc = kine.changeMarkersReferenceFrame(targetMarkers, RT1i)
+        
+        # SVD for acquisition system 2
+        clusterBaseMarkerLoc = clusterInfoSpec[clusterBaseNameSpec]
+        args = {}
+        markersLoc = clusterInfoSpec.copy()
+        del markersLoc[clusterBaseNameSpec]
+        args['mkrsLoc'] = markersLoc
+        args['verbose'] = True
+        mkrList = markersLoc.keys()
+        R2, T2, infoSVD2 = kine.rigidBodySVDFun2(markers2, mkrList, args)
+        RT2 = kine.composeRotoTranslMatrix(R2, T2)
+        
+        # Copy and rename cluster base marker to be expressed in system 2 in global reference frame
+        targetMarkersLoc[clusterBaseNameSpec + ' 2'] = clusterBaseMarkerLoc
+        
+        # Update list of markers for system 2 in global reference frame
+        markers2New.update(kine.changeMarkersReferenceFrame(targetMarkersLoc, RT2))
+    
+    # Express spine points in pelvis reference frame
+    spinePoints = {m: markers2New[m] for m in spinePointNames}
+    spinePointsPel = kine.changeMarkersReferenceFrame(spinePoints, RTPel2i)
+    
+    # Merge spine points in one array
+    spineData = np.stack([spinePointsPel[m] for m in spinePointNames], axis=2)  # Nf x 3 x Np
+    
+    # Init results
+    res = {}
+    res['newMarkers'] = markers2New
+    sup, inf = spinePointNames[:-1], spinePointNames[1:]
+    spineAngleNames = [sup[i] + '_' + inf[i] for i in range(len(sup))]
+    Nf = spineData.shape[0]
+    res['spineAngles'] = {}
+    res['spineAngles']['sagittal'] = {a: np.zeros((Nf,)) for a in spineAngleNames}
+    
+    # Create results directory if not existing
+    if not os.path.exists(resultsDir):
+        os.mkdir(resultsDir)
+    
+    # Process data
+    for i in range(Nf):
+        
+        print('processing time frame %d ...' % i)        
+        
+        # Interpolate spline in sagittal plane
+        spineDataSag = spineData[i,0:2,:].T # Np x 2
+        Np = spineDataSag.shape[0]
+        spineLineSag = spine.create2DSpline(spineDataSag)
+        
+        # Calculate slope of spine normal at the wanted points
+        spineLineSagDer = spine.calcSplineTangentSlopes(spineDataSag, u='only_pts')
+        normalSlopes = -spineLineSagDer[:,0] / spineLineSagDer[:,1]
+
+        # Calculate angles between segments
+        m1, m2 = normalSlopes[:-1], normalSlopes[1:]
+        angles = spine.calcInterlinesAngle(m1, m2)
+        for j in range(len(spineAngleNames)):
+            res['spineAngles']['sagittal'][spineAngleNames[j]][i] = angles[j]
+        
+        if savePlots:
+            
+            # Create results directory if not existing
+            figuresDir = os.path.join(resultsDir, 'figures')
+            if not os.path.exists(figuresDir):
+                os.mkdir(figuresDir)
+            
+            # Plot spine in 3D
+#            fig = plt.figure()
+#            ax = fig.gca(projection='3d')
+#            ax.scatter(spineData[i,2,:], spineData[i,0,:], spineData[i,1,:])
+#            mplh.set_axes_equal(ax)          
+            
+            # Plot spine in sagittal/frontal plane
+            fig = plt.figure()
+            
+            plt.subplot(1, 2, 1)
+            plt.plot(spineDataSag[:,0], spineDataSag[:,1], 'o')
+            plt.plot(spineLineSag[:,0], spineLineSag[:,1], lw=3)
+            ax = plt.gca()
+            xlim, ylim = ax.get_xlim(), ax.get_ylim()
+            xN = np.tile([-1000, 1000], [Np, 1])
+            yN = (xN - spineDataSag[:,:1]) * normalSlopes[:,None] + spineDataSag[:,1:2]
+            plt.plot(xN.T, yN.T, 'k')
+            plt.axis('equal')
+            ax.set_xlim(xlim)
+            ax.set_ylim(ylim)
+            plt.title('Sagittal')
+            plt.xlabel('X pelvis (anterior +)')
+            plt.ylabel('Y pelvis (up +)')
+            
+            plt.subplot(1, 2, 2)
+            plt.plot(spineData[i,2,:], spineData[i,1,:], 'o')
+            plt.axis('equal')
+            plt.title('Frontal')
+            plt.xlabel('Z pelvis (right +)')
+            plt.savefig(os.path.join(figuresDir, 'tf_%04d.png' % i), format='png', orientation='landscape')
+    
+    # Create MATLAB-friendly reslts structure
+    def adjustDictForML(d):
+        dML = {}
+        for k in d:
+            kML = k.replace(' ', '_')
+            dML[kML] = d[k]
+        return dML
+    resML = res.copy()
+    resML['newMarkers'] = adjustDictForML(res['newMarkers'])        
+    resML['spineAngles']['sagittal'] = adjustDictForML(res['spineAngles']['sagittal'])
+    
+    # Save data to MAT file
+    fio.writeMATFile(os.path.join(resultsDir, 'results.mat'), resML)
+        
+    return res
+    
+    
+    
+
+    
     
\ No newline at end of file
